from django.shortcuts import render

from objects.dashboards.home.dashboard import Dashboard as HomeDashboard
from reporting.lib import get_entity_by_class, get_reports_list


# TODO: добавить декоратор @login_required() к вьюхе, которая недоступна незалогиненным пользователям
# from django.contrib.auth.decorators import login_required


def index(request):
    # на главной странице выводится дашборд home
    dashboard = HomeDashboard(request.GET)

    return render(request, 'reporting/dashboard_detail.html', {'dashboard': dashboard})


def report_list(request):
    reports = get_reports_list()

    context = {'report_list': reports}

    return render(request, 'reporting/list.html', context)


def report_detail(request, report_id):
    report = get_entity_by_class('objects.reports.{}.report'.format(report_id), 'Report', request.GET)

    return render(request, 'reporting/report_detail.html', {'report': report})


def report_detail_raw(request, report_id):
    report = get_entity_by_class('objects.reports.{}.report'.format(report_id), 'Report', request.GET)

    return report.get_data()


def dashboard_detail(request, dashboard_id):
    dashboard = get_entity_by_class('objects.dashboards.{}.dashboard'.format(dashboard_id), 'Dashboard', request.GET)

    return render(request, 'reporting/dashboard_detail.html', {'dashboard': dashboard})


def dashboard_detail_nested(request, dashboard_id, dashboard_parent_id):
    dashboard = get_entity_by_class(
        'objects.dashboards.{}.{}.dashboard'.format(dashboard_parent_id, dashboard_id), 'Dashboard', request.GET)

    return render(request, 'reporting/dashboard_detail.html', {'dashboard': dashboard})



from django.http import QueryDict
from django.template import Library, loader

register = Library()


@register.simple_tag
def report(report_class):
    rc = __import__(report_class, globals(), locals(), ['*'])
    cls = getattr(rc, 'Report')
    re = cls(QueryDict())

    context = {'report': re}
    template = loader.get_template('reporting/dashboard_report.html')
    return template.render(context)



from django.test import TestCase

from reporting.lib import transform_python_list_to_list_for_echarts, get_entity_by_class, get_dashboards_hierarchy, \
    get_dashboards_hierarchy_for_template, convert_dashboard_class_to_tuple, get_reports_list
from reporting.tests.fixtures.objects.dashboards.dummy1.dashboard import Dashboard as DummyBoard1
from reporting.tests.fixtures.objects.dashboards.dummy1.dummy3.dashboard import Dashboard as DummyBoard3
from reporting.tests.fixtures.objects.dashboards.dummy2.dashboard import Dashboard as DummyBoard2
from reporting.tests.fixtures.objects.reports.dummy1.report import Report as DummyReport1
from reporting.tests.fixtures.objects.reports.dummy2.report import Report as DummyReport2


class LibTests(TestCase):
    def test_transform_python_list_to_list_for_echarts(self):
        self.assertEqual(transform_python_list_to_list_for_echarts([1, 2, 3]), "['1', '2', '3']")

    def test_get_entity_by_class(self):
        entity = get_entity_by_class('reporting.tests.fixtures.objects.reports.dummy1.report', 'Report', {})
        self.assertEqual(type(entity), DummyReport1)

    def test_get_dashboards_hierarchy(self):
        self.assertEqual(
            {DummyBoard1: [DummyBoard3], DummyBoard2: []},
            get_dashboards_hierarchy('reporting/tests/fixtures/'))

    def test_get_report_list(self):
        self.assertEqual([type(item) for item in get_reports_list('reporting/tests/fixtures/')],
                         [DummyReport1, DummyReport2])

    def test_dashboards_hierarchy_for_template(self):
        self.assertEqual(get_dashboards_hierarchy_for_template('reporting/tests/fixtures/'),
                         {('dummy1', 'Dummy board 1', 'fa fa-pie-chart', None): [
                             ('dummy3', 'Dummy board 3', 'fa fa-pie-chart', 'dummy1')],
                             ('dummy2', 'Dummy board 2', 'fa fa-pie-chart', None): []}
                         )

    def test_convert_dashboard_class_to_tuple(self):
        self.assertEqual(convert_dashboard_class_to_tuple(DummyBoard1),
                         ('dummy1', 'Dummy board 1', 'fa fa-pie-chart', None))



import glob
import hashlib
import os
import sys
from typing import Union, Dict, List, Tuple, Type

from django.core.cache import cache
from django.http import Http404, QueryDict

from reporting.models.dashboard import BaseDashboard
from reporting.models.report import BaseReport


def transform_python_list_to_list_for_echarts(l: list) -> str:
    """Преобразует питоновский лист в строку вида '['abc', 'efg']' для echarts.

    :param l: список, который нужно преобразовать
    :return: строка для echarts
    """
    return '[\'' + '\', \''.join([str(i) for i in l]) + '\']'


def get_entity_by_class(path: str, class_name: str, class_params: dict = None) -> Union[
    BaseReport, BaseDashboard]:
    """Возвращает экземпляр класса по пути до файла и имени класса.

    :param class_params: параметры класса отчёта
    :param path: путь до файла класса (например, objects.dashboards.dummydashboard.dashboard
    :param class_name: название класса
    :return:
    """
    try:
        module = __import__(path, globals(), locals(), ['*'])

        entity_class = getattr(module, class_name)
        return entity_class(params=class_params)
    except ModuleNotFoundError:
        # если такой репорт не найден
        # TODO: не хорошо, что тут что-то про 404, так как тут вообще может быть не про веб
        raise Http404()


def get_class_by_class_path(path: str, class_name: str, class_params: dict = None):
    try:
        module = __import__(path, globals(), locals(), ['*'])

        entity_class = getattr(module, class_name)
        return entity_class
    except ModuleNotFoundError:
        # если такой репорт не найден
        # TODO: не хорошо, что тут что-то про 404, так как тут вообще может быть не про веб
        raise Http404()


def get_reports_list(path_to_objects='') -> List[Type[BaseReport]]:
    """Возвращает список экземпляров отчётов.

    :param path_to_objects:
    :return:
    """
    reports_list = []
    files = glob.iglob(os.path.join(path_to_objects, 'objects', 'reports', '**', 'report.py'), recursive=True)
    files = list(files)
    for file in sorted(files):
        paths = file.split('/')
        paths[-1] = paths[-1][0:-3]
        module_name = '.'.join(paths)

        __import__(module_name, globals(), locals(), ['*'])
        cls = getattr(sys.modules[module_name], 'Report')

        reports_list.append(cls(QueryDict()))
    return reports_list


def get_dashboards_hierarchy(path_to_objects='') -> Dict[Type[BaseDashboard], List[Type[BaseDashboard]]]:
    """Возвращает иерархию классов дашбордов.

    :param path_to_objects: путь до директории с объектами (по умолчанию - пустая строка, директория в корне проекта)
    :return:
    """

    def get_len_of_path_to_objects(path_to_objects_dir=''):
        return len(path_to_objects_dir.split('/')) - 1

    dashboards_hierarchy = {}
    files = glob.iglob(os.path.join(path_to_objects, 'objects', 'dashboards', '**', 'dashboard.py'), recursive=True)
    files = list(files)
    for file in sorted(files):
        paths = file.split('/')
        paths[-1] = paths[-1][0:-3]
        module_name = '.'.join(paths)

        __import__(module_name, globals(), locals(), ['*'])
        cls = getattr(sys.modules[module_name], 'Dashboard')

        if cls not in dashboards_hierarchy and len(paths) == get_len_of_path_to_objects(path_to_objects) + 4:
            dashboards_hierarchy[cls] = []
        if len(paths) == get_len_of_path_to_objects(path_to_objects) + 5:
            if cls.get_parent_dashboard_class() not in dashboards_hierarchy.keys():
                dashboards_hierarchy[cls.get_parent_dashboard_class()] = [cls]
            else:
                dashboards_hierarchy[cls.get_parent_dashboard_class()].append(cls)

    return dashboards_hierarchy


def convert_dashboard_class_to_tuple(dashboard_class: Type[BaseDashboard]) -> Tuple:
    """Преобразует класс дашборда в тупл для использования в шаблонах.

    :param dashboard_class:
    :return:
    """
    board = dashboard_class(QueryDict())
    result = [board.id,
              board.title,
              board.icon,
              dashboard_class.get_parent_dashboard_id()]
    return tuple(result)


def get_dashboards_hierarchy_for_template(path_to_objects='') -> dict:
    """Возвращает иерархию дашбордов в виде словаря туплов.
    Для чего это ... сделано: в темплейтах классы автоматически инстанцируются, поэтому сделано на туплах

    :param path_to_objects:
    :return:
    """
    dashboards_hierarchy_class = get_dashboards_hierarchy(path_to_objects)
    dashboards_hierarchy_for_template = {}

    for dashboards_hierarchy_class_key in dashboards_hierarchy_class.keys():
        temp_key = convert_dashboard_class_to_tuple(dashboards_hierarchy_class_key)
        dashboards_hierarchy_for_template[temp_key] = []
        for dashboards_hierarchy_class_value in dashboards_hierarchy_class[dashboards_hierarchy_class_key]:
            dashboards_hierarchy_for_template[temp_key].append(
                convert_dashboard_class_to_tuple(dashboards_hierarchy_class_value))
    return dashboards_hierarchy_for_template


def cache_dataframe(fn):
    """
    Декоратор для кеширования датафрейма в методе get_dataframe датасета.

    :param fn:
    :return:
    """
    CACHE_TIMOUT = 1 * 7 * 24 * 60 * 60  # неделя

    def cache_get_key(*args):
        serialise = []
        for arg in args:
            serialise.append(str(arg))

        full_str = ''.join(serialise).encode('utf-8')
        key = hashlib.md5(full_str).hexdigest()
        return key

    def memoized_func(dataset, params=None):
        _cache_key = cache_get_key(fn.__name__, type(dataset), params)
        result = cache.get(_cache_key)
        if result is None:
            result = fn(dataset, params)
            cache.set(_cache_key, result, CACHE_TIMOUT)
        return result

    return memoized_func



from abc import ABC, abstractmethod
from typing import Dict

import pandas as pd

from reporting.lib import cache_dataframe


class BaseDataset(ABC):
    """
    Абстрактный класс для всех классов датасетов (они должны быть от него отнаследованы).
    """

    @abstractmethod
    def get_dataframe(self, params: Dict = None) -> pd.DataFrame:
        """
        Возвращает данные по оси x и y (возможно несколько).

        :return:
        """
        pass

    @cache_dataframe
    def get_cached_dataframe(self, params: Dict = None) -> pd.DataFrame:
        """
        Возвращает закешированный датафрейм.

        :param params:
        :return:
        """
        # TODO: возможно лучше живьём обернуть, а не декоратором ... хз
        return self.get_dataframe(params)



from .lib import get_dashboards_hierarchy_for_template


def add_variable_to_context(request):
    # получаем список дашбордов
    dashboards_hierarchy = get_dashboards_hierarchy_for_template()

    return {
        'dashboards_hierarchy': dashboards_hierarchy
    }



import hashlib
import urllib.parse

from django import template
from django.utils.safestring import mark_safe

register = template.Library()


@register.filter
def gravatar_url(email, size=40):
    """
    Возвращает только урл граватара.

    Пример использования: {{ email|gravatar_url:150 }}

    :param email:
    :param size:
    :return:
    """
    return "https://www.gravatar.com/avatar/%s?%s" % (
        hashlib.md5(email.lower().encode('utf-8')).hexdigest(), urllib.parse.urlencode({'s': str(size)}))


@register.filter
def gravatar(email, size=40):
    """
    Возвращает image tag граватара.

    Пример использования:  {{ email|gravatar:150 }}

    :param email:
    :param size:
    :return:
    """
    url = gravatar_url(email, size)
    return mark_safe('<img src="%s" height="%d" width="%d">' % (url, size, size))



from abc import ABC, abstractmethod
from typing import Dict, Type

from django.forms import Form
from django.http import QueryDict
from django.urls import reverse


class BaseReport(ABC):
    """Абстрактный класс для всех классов отчётов (они должны быть от него отнаследованы).

        form_class  Определеине класса формы
        form_defaults   Значения полей формы по умолчанию
    """

    _params: QueryDict = None

    form_class: Type[Form] = None
    form_defaults: Dict = {}

    def __init__(self, params: QueryDict) -> None:
        """Констурктор.

        :param params: параметры отчета
        """
        self._params = params

    @property
    def id(self) -> str:
        """Возвращает идентификатор отчёта.

        :return:
        """
        return str(self.__class__.__module__).split('.')[-2]

    @property
    @abstractmethod
    def title(self) -> str:
        """заголовок отчёта

        :return:
        """
        pass

    @property
    @abstractmethod
    def description(self) -> str:
        """описание отчёта

        :return:
        """
        pass

    @property
    def template(self) -> str:
        """возвращает путь до темплейта отчёта

        :return:
        """
        return 'reports/{}/template.html'.format(self.id)

    @property
    def container_id(self) -> str:
        """возвращает идентификатор div, в котором будет отрисован график

        :return:
        """
        return '{}_report'.format(self.id)

    def get_form(self) -> Form:
        """Возвращате экземпляр формы отчета

        :return:
        """
        if self.has_form():
            params = QueryDict(mutable=True)
            params.update(self.form_defaults)
            params.update(self._params)

            form = self.form_class(params)
            form.is_valid()

            return form

    def has_form(self) -> bool:
        """Сущесвование формы

        :return:
        """
        return self.form_class is not None

    def get_raw_view_url(self) -> str:
        """URL до raw entry point

        :return:
        """
        return reverse('reporting:report-detail-raw', args=[self.id]) + '?' + self._params.urlencode()



from abc import ABC, abstractmethod
from typing import Union, Dict, Type, Text

from django.forms import Form
from django.http import QueryDict


# TODO: сделать общий абстрактный класс с репортом, в который вынести тайтл и дескрипшн

class BaseDashboard(ABC):
    """
    Базовый класс для всех классов дашбордов (они должны быть от него отнаследованы).

        form_class  Определеине класса формы
        form_defaults   Значения полей формы по умолчанию
    """

    _params: QueryDict = None
    form_class: Type[Form] = None
    form_defaults: Dict = {}

    # TODO: проверить целесоорбразность параметра парамс
    def __init__(self, params: QueryDict):
        """
        Констурктор.

        :param params: параметры отчета
        """
        self._params = params
        pass

    @property
    def id(self) -> Text:
        """
        Идентификатор дашборда.

        :return:
        """
        return str(self.__module__).split('.')[-2]

    @property
    def icon(self) -> Text:
        return "fa fa-pie-chart"

    @property
    @abstractmethod
    def title(self) -> Text:
        """
        Заголовок дашборда.

        :return:
        """
        pass

    @property
    def template(self) -> Text:
        """
        Путь до темплейта дашборда.

        :return:
        """
        if self.get_parent_dashboard_id():
            return 'dashboards/{}/{}/template.html'.format(self.get_parent_dashboard_id(), self.id)
        else:
            return 'dashboards/{}/template.html'.format(self.id)

    @classmethod
    def get_parent_dashboard_id(cls) -> Union[Text, None]:
        module_splitted = cls.__module__.split('.')
        if module_splitted[-3] == 'dashboards':
            return None
        else:
            return module_splitted[-3]

    @classmethod
    def get_parent_dashboard_class(cls):
        module_splitted = cls.__module__.split('.')
        module = __import__('.'.join(module_splitted[:-2]) + '.dashboard', globals(), locals(), ['*'])

        entity_class = getattr(module, 'Dashboard')
        return entity_class

    def get_form(self) -> Form:
        """
        Возвращате экземпляр формы отчета.

        :return:
        """
        if self.has_form():
            params = QueryDict(mutable=True)
            params.update(self.form_defaults)
            params.update(self._params)

            form = self.form_class(params)
            form.is_valid()

            return form

    def has_form(self) -> bool:
        """
        Сущесвование формы.

        :return:
        """
        return self.form_class is not None



from django.http import QueryDict
from django.test import TestCase

from reporting.tests.fixtures.objects.dashboards.dummy1.dashboard import Dashboard as DummyBoard1
from reporting.tests.fixtures.objects.dashboards.dummy1.dummy3.dashboard import Dashboard as DummyBoard3
from reporting.tests.fixtures.objects.reports.dummy1.report import Report as DummyReport1


class DashboardTests(TestCase):
    def test_id(self):
        board = DummyBoard1(QueryDict())
        self.assertEqual(board.id, 'dummy1')

    def test_template(self):
        board = DummyBoard1(QueryDict())
        self.assertEqual(board.template, 'dashboards/dummy1/template.html')

    def test_get_parent_dashboard_id(self):
        self.assertEqual(DummyBoard3.get_parent_dashboard_id(), 'dummy1')
        self.assertIsNone(DummyBoard1.get_parent_dashboard_id())

    def test_get_parent_dashboard_class(self):
        self.assertEqual(DummyBoard3.get_parent_dashboard_class(), DummyBoard1)


class ReportTests(TestCase):
    def test_id(self):
        dr1 = DummyReport1(QueryDict())
        self.assertEqual(dr1.id, 'dummy1')

    def test_template(self):
        dr1 = DummyReport1(QueryDict())
        self.assertEqual(dr1.template, 'reports/dummy1/template.html')

    def test_container_id(self):
        dr1 = DummyReport1(QueryDict())
        self.assertEqual(dr1.container_id, 'dummy1_report')
